#!/usr/bin/env python

#
# COMPUTE_GB_TILT_ENERGY.PY
#
# Author:      Brandon Runnels
# Date:        June 2017
# Institution: Department of Mechanical and Aerospace Engineering
#              University of Colorado Colorado Springs
# Email:       brunnels@uccs.edu
#
# Description: This file contains a single function that computes
#              the grain boundary (GB) energy for a range of tilt
#              angles and an arbitrarily selected tilt axis.
#              The user specifies the initial orientations and the
#              range of tilt angles, as well as a maximum repeat-
#              ing unit cell size. The script then creates a 
#              LAMMPS input file for all possible tilt boundaries
#              within that set of parameters, runs LAMMPS, 
#              collects output, and returns the theta ranges, en-
#              ergies, sigma values, and minimum seperation dist-
#              ances to the calling function.
#              For details on the input parameters, please see
#              README.txt
#
# Copyright:   2016 University of Colorado Colorado Springs
#

import numpy
from scipy import linalg
from fractions import Fraction
from math import ceil
import os,os.path,sys,re

def compute_gb_tilt_energy(
        output_dir,
        a1, 
        a2, 
        a3,
        pair_style,
        pair_coeff,
        theta_min,
        theta_max,
        max_denominator,
        lattice_style,
        lattice_constant,
        cutoff_distance,
        x_repeat = 1,
        z_repeat = 1,
        min_cell_height = 10,
        offsets_top = -1,
        offsets_bottom = -1,
        num_mpi_processors = 1,
        cohesive_energy = 0,
        mass = 0,
        gzip=False,
        verbose=False):

    #lammps_path = '/opt/lammps/src/lmp_serial'
    lammps_path = 'lammps'


    grid_search=False
    if offsets_top == -1 and offsets_bottom == -1:
        grid_search=True
    elif offsets_top == -1 or offsets_bottom == -1:
        raise("Error: you must specify BOTH offsets_top and offsets_bottom or NEITHER offsets_top or offsets_bottom")
        

    theta_res = 1000

    # unit vectors corresponding to the un-tilted configuration
    a1 = numpy.asarray(a1)
    a2 = numpy.asarray(a2)
    a3 = numpy.asarray(a3)
    e1 = a1/linalg.norm(a1)
    e2 = a2/linalg.norm(a2)
    e3 = a3/linalg.norm(a3)
    # check for orthogonality
    if (abs(a1.dot(a2)) > 1E-8 or abs(a2.dot(a3)) > 1E-8 or abs(a3.dot(a1)) > 1E-8): 
        raise Exception("Non-orthogonal vectors")
    # check for right hand rule
    if (numpy.cross(e1,e2).dot(e3) < 0 or numpy.cross(e2,e3).dot(e1) < 0 or numpy.cross(e3,e1).dot(e2) < 0): 
        raise Exception("Vectors do not obey right hand rule")

    # initialize output arrays
    thetas           =[]
    gb_energies      =[]
    minimum_distances=[]
    sigmas           =[]

    # Integers to remember the last CSL configuration to avoid repeating calculations
    mold=0;nold=0;pold=0;qold=0

    # Iterate thorough all thetas in the range
    for th in numpy.linspace(theta_min,theta_max,theta_res):
        # Compute the repeating distance in the x direction and snap to the nearest finite integer rationalization
        if -1 < (numpy.tan(numpy.radians(th/2))*linalg.norm(a1)/linalg.norm(a2)) < 1:
            frac = Fraction(numpy.tan(numpy.radians(th/2))*linalg.norm(a1)/linalg.norm(a2)).limit_denominator(max_denominator);
            m = int(frac.denominator); n = int(frac.numerator); 
        else:
            frac = Fraction(linalg.norm(a2)/numpy.tan(numpy.radians(th/2))/linalg.norm(a1)).limit_denominator(max_denominator);
            m = int(frac.numerator); n = int(frac.denominator);
        
        # Compute the repeating distance in the y direction and snap to the nearest finite integer rationalization
        if -1 < (numpy.tan(numpy.radians(th/2))*linalg.norm(a2)/linalg.norm(a1)) < 1:
            frac = Fraction(numpy.tan(numpy.radians(th/2))*linalg.norm(a2)/linalg.norm(a1)).limit_denominator(max_denominator);
            p = int(frac.numerator); q = int(frac.denominator); 
        else:
            frac = Fraction(linalg.norm(a1)/numpy.tan(numpy.radians(th/2))/linalg.norm(a2)).limit_denominator(max_denominator);
            p = int(frac.denominator); q = int(frac.numerator); 
    
        # Avoid repeating previously found rationalizations
        if [m,n,p,q]==[mold,nold,pold,qold]: continue
        mold=m;nold=n;pold=p;qold=q;

        # Find the actual theta for the rationalized angle
        thetaA = numpy.degrees(2*numpy.arctan2(n*linalg.norm(a2),m*linalg.norm(a1)))
        thetaB = numpy.degrees(2*numpy.arctan2(p*linalg.norm(a1),q*linalg.norm(a2)))

        # Sometimes the top and bottom crystals return different angles; in this case, skip
        if abs(thetaA-thetaB) > 1E-8: continue
        theta = thetaA

        if verbose:
            print("\n========================================");
            print("theta = "+str(theta));
            print("========================================\n");

        # Compute integral dimensions of the unit cell
        lambda1 = m*linalg.norm(a1)*numpy.cos(numpy.radians(theta/2)) + n*linalg.norm(a2)*numpy.sin(numpy.radians(theta/2))
        lambda2 = p*linalg.norm(a1)*numpy.sin(numpy.radians(theta/2)) + q*linalg.norm(a2)*numpy.cos(numpy.radians(theta/2))
        lambda3 = linalg.norm(a3)

        # Convert integral dimensions to actual lengths
        x_cell = lambda1*lattice_constant
        y_cell = lambda2*lattice_constant
        z_cell = lambda3*lattice_constant

        # Compute the minimum number of repeats in the y direction to satisfy the minimum height requirement
        y_repeat = int(ceil(min_cell_height*lattice_constant/y_cell))

        # Compute coordinates of the box
        xmin = 0
        xmax = x_cell * x_repeat;
        ymin = -y_cell * y_repeat  - 0.001*lattice_constant;
        ymax = y_cell * y_repeat  + 0.001*lattice_constant;
        zmin = 0;
        zmax = z_cell * z_repeat;

        # Compute rotated direction vectors
        X1 = m*a1 - n*a2;  X2 = m*a1 + n*a2
        Y1 = p*a1 + q*a2;  Y2 = -p*a1 + q*a2
        Z1 = a3;           Z2 = a3
        
        num_points_x = 10;
        num_points_z = 3;
        if grid_search == True:
            x_shifts = [n*lambda1/num_points_x for n in range(0,num_points_x)]
            z_shifts = [n*lambda3/num_points_z for n in range(0,num_points_z)]
            offsets_top = [[0.5*x,0,0.5*z] for x in x_shifts for z in z_shifts]
            offsets_bottom = [[-0.5*x,0,-0.5*z] for x in x_shifts for z in z_shifts]
        
        minimum_energy = float('inf')
        for offset_top, offset_bottom in zip(offsets_top,offsets_bottom):
            # Create the directory structure for dumping the output
            directory = output_dir + '/dump_'+format(theta, '.4f').zfill(8)+'/'
            #os.system('rm -rf ' + directory)
            if not os.path.exists(directory): os.makedirs(directory)
            #if(test_var): exit(0);

            # Create LAMMPS input file
            file = open(directory + 'generated_input.gb','w')

            # Simulation set up
            file.write('clear              \n')
            file.write('units              metal \n')
            file.write('dimension          3 \n')
            file.write('boundary           p p p \n')
            file.write('atom_style         atomic\n')
            # Simulation domain
            file.write('lattice            ' + lattice_style + ' ' + str(lattice_constant) + ' \n')
            file.write('region             whole block ' + str(xmin) + ' ' + str(xmax) + ' ' + str(ymin) + ' ' + str(ymax) + ' ' + str(zmin) + ' ' + str(zmax) + ' units box \n')
            file.write('create_box         2 whole\n')
            # Upper region
            file.write('region             upper block INF INF 0.0 ' + ' ' + str(ymax) + ' INF INF units box \n')
            file.write('lattice            ' + lattice_style + ' ' + str(lattice_constant) 
                       # + ' origin ' + ' '.join(map(str,offset_top)) # superseded by displace_atoms command
                       + ' orient x ' + ' '.join(map(str,X1)) 
                       + ' orient y ' + ' '.join(map(str,Y1)) 
                       + ' orient z ' + ' '.join(map(str,Z1)) + ' \n')
            file.write('create_atoms       1 region upper \n')
            file.write('group              upper type 1\n')
            if abs(mass)>1E-10:
                file.write('mass               1 '+str(mass)+'\n')
            # Move atoms in upper region
            file.write('displace_atoms     upper move ' + ' '.join(map(str,offset_top)) + ' \n')
            # Lower region
            file.write('region             lower block INF INF ' + str(ymin) + ' 0.0 INF INF units box \n')
            file.write('lattice            ' + lattice_style + ' ' + str(lattice_constant) 
                       #+ ' origin ' + ' '.join(map(str,offset_bottom)) # superseded by displace_atoms command
                       + ' orient x ' + ' '.join(map(str,X2)) 
                       + ' orient y ' + ' '.join(map(str,Y2)) 
                       + ' orient z ' + ' '.join(map(str,Z2)) + ' \n')
            file.write('create_atoms       2 region lower \n')
            file.write('group              lower type 2\n')
            # Move atoms in lower region
            file.write('displace_atoms     lower move ' + ' '.join(map(str,offset_bottom)) + ' \n')
            if abs(mass)>1E-10:
                file.write('mass               2 '+str(mass)+'\n')
            # Interatomic potential and neighbor settings
            file.write('pair_style         '+pair_style+'\n')
            file.write('pair_coeff         '+pair_coeff+'\n')
            file.write('neighbor           2.0 bin \n')
            file.write('neigh_modify       delay 10 check yes\n')
            # Delete overlapping atoms
            file.write('delete_atoms       overlap ' + str(cutoff_distance*lattice_constant) + ' all all \n')
            # Solver
            file.write('min_style          cg  \n')
            file.write('minimize           1e-15 1e-15 5000 10000  \n')
            file.write('fix                1 all box/relax x 0.0 z 0.0 couple none vmax 0.001 \n') 
            file.write('minimize           1.0e-25 1.0e-25 1000 10000 \n') 
            file.write('unfix              1 \n') 
            # Compute centrosymmetry, energy, and min dist variables
            file.write('compute            csym       all     centro/atom '+lattice_style+' \n')
            file.write('compute            csymsum    all     reduce sum c_csym \n')
            file.write('compute            eng        all     pe/atom \n')
            file.write('compute            eatoms     all     reduce sum c_eng \n')
            file.write('compute            distance   all     pair/local dist \n')
            file.write('compute            mindist    all     reduce min c_distance \n')
            file.write('reset_timestep     0  \n')
            file.write('run                0  \n')
            file.write('thermo             0  \n')
            file.write('thermo_style       custom step pe lx ly lz press pxx pyy pzz c_eatoms c_mindist c_csymsum  \n')
            file.write('run                0  \n')
            # Spit out total energy
            file.write('variable           total_energy equal c_eatoms\n')
            file.write('print              "${total_energy}" file ' + directory + 'energy.out\n')
            # Spit out number of atoms
            file.write('variable           num_atoms equal "count(all)" \n')
            file.write('print              "${num_atoms}" file ' + directory + 'numatoms.out\n')
            # Spit out distances
            file.write('variable           minimum_distance equal c_mindist\n')
            file.write('print              "${minimum_distance}" file ' + directory + 'mindistance.out\n')
            # Write out dump file
            file.write('write_dump         all cfg' + ('/gz ' if gzip else ' ') + directory + 'dump_post.cfg' + ('.gz ' if gzip else ' ') + 'mass type xs ys zs id type xs ys zs c_csym c_eng modify element Al Al \n')
            file.write('print              "This indicates that LAMMPS ran successfully" file ' + directory + 'success.out')
            file.close()

            # Run LAMMPS
            if num_mpi_processors > 1:
                os.system('mpirun -np ' + str(num_mpi_processors) + ' lammps < ' + directory + 'generated_input.gb > ' + directory + 'run_output.gb -log ' + directory + 'lammps.out')
            elif num_mpi_processors == 1:
                os.system(lammps_path + ' < ' + directory + 'generated_input.gb > ' + directory + 'run_output.gb -log ' + directory + 'lammps.out')
            else:
                raise Exception("Error: invalid number of processors")
                
            ###if not os.path.isfile(directory+'dump_post.'+dump_format+('.gz ' if gzip else '')):
            if not os.path.isfile(directory+'dump_post.cfg'+('.gz ' if gzip else '')):
                raise Exception("Error: LAMMPS did not generate a dump file -- something probably went wrong.")

            if not os.path.isfile(directory+'success.out'):
                raise Exception("Error: there was an error in the LAMMPS run -- check the LAMMPS log file for more detail")

            # Collect output from LAMMPS that was recorded in files
            energy      = float(open(directory+'energy.out', 'r').read())
            numatoms    = int(open(directory+'numatoms.out', 'r').read())
            mindistance = float(open(directory+'mindistance.out', 'r').read())

            # Compute grain boundary energy
            if theta==0 and abs(cohesive_energy)<1E-16: cohesive_energy = energy/numatoms;
            excess_energy         = energy - numatoms*(-abs(cohesive_energy))
            gb_energy_ev_anstrom2 = excess_energy / (x_repeat*(xmax-xmin)) / (z_repeat*(zmax-zmin)) / 2
            gb_energy_J_m2        = gb_energy_ev_anstrom2*16.0218

            print("Energy: " + str(gb_energy_J_m2));
            
            if gb_energy_J_m2 < minimum_energy:
                test_var = True;
                minimum_energy = gb_energy_J_m2;
                os.renames(directory+"dump_post.cfg",directory+"dump_post_min.cfg");
                os.renames(directory+"generated_input.gb",directory+"generated_input_min.gb");
                min_engr_file = open(directory+'energy_min.out',"w")
                min_engr_file.write(str(minimum_energy));
                min_engr_file.close();
            
        # Compute sigma value
        sigma = int(lambda1*lambda2*lambda3)
        while sigma%2 == 0: sigma = sigma/2
        
        # Store results
        thetas.append(theta)
        if verbose: print("   minimum gb energy = " + str(minimum_energy))
        gb_energies.append(minimum_energy)        
        if verbose: print("   minimum distance = " + str(mindistance))
        minimum_distances.append(mindistance)
        if verbose: print("   sigma value = " + str(sigma))
        sigmas.append(sigma)

    return([thetas, gb_energies,minimum_distances,sigmas]);
        
